@@ -7,9 +7,12 @@
  *   (at your option) any later version.                                   *
  ***************************************************************************/
 
+#include "pageviewutils.h"
+
 // qt/kde includes
 #include <qapplication.h>
 #include <qbitmap.h>
+#include <qbrush.h>
 #include <qimage.h>
 #include <qlabel.h>
 #include <qlayout.h>
@@ -19,16 +22,16 @@
 #include <qtimer.h>
 #include <qtoolbutton.h>
 #include <kacceleratormanager.h>
-#include <kicon.h>
+#include <kcolorscheme.h>
 #include <kiconloader.h>
-#include <kimageeffect.h>
 #include <klocale.h>
 
 // system includes
 #include <math.h>
 
 // local includes
-#include "pageviewutils.h"
+#include "formwidgets.h"
+#include "guiutils.h"
 #include "core/page.h"
 #include "settings.h"
 
@@ -37,10 +40,18 @@
 /*********************/
 
 PageViewItem::PageViewItem( const Okular::Page * page )
-    : m_page( page ), m_zoomFactor( 1.0 )
+    : m_page( page ), m_zoomFactor( 1.0 ), m_visible( true ),
+    m_formsVisible( false ), m_crop( 0., 0., 1., 1. )
 {
 }
 
+PageViewItem::~PageViewItem()
+{
+    QHash<int, FormWidgetIface*>::iterator it = m_formWidgets.begin(), itEnd = m_formWidgets.end();
+    for ( ; it != itEnd; ++it )
+        delete *it;
+}
+
 const Okular::Page * PageViewItem::page() const
 {
     return m_page;
@@ -51,19 +62,39 @@ int PageViewItem::pageNumber() const
     return m_page->number();
 }
 
-const QRect& PageViewItem::geometry() const
+const QRect& PageViewItem::croppedGeometry() const
 {
-    return m_geometry;
+    return m_croppedGeometry;
 }
 
-int PageViewItem::width() const
+int PageViewItem::croppedWidth() const
 {
-    return m_geometry.width();
+    return m_croppedGeometry.width();
 }
 
-int PageViewItem::height() const
+int PageViewItem::croppedHeight() const
 {
-    return m_geometry.height();
+    return m_croppedGeometry.height();
+}
+
+const QRect& PageViewItem::uncroppedGeometry() const
+{
+    return m_uncroppedGeometry;
+}
+
+int PageViewItem::uncroppedWidth() const
+{
+    return m_uncroppedGeometry.width();
+}
+
+int PageViewItem::uncroppedHeight() const
+{
+    return m_uncroppedGeometry.height();
+}
+
+const Okular::NormalizedRect & PageViewItem::crop() const
+{
+    return m_crop;
 }
 
 double PageViewItem::zoomFactor() const
@@ -71,27 +102,88 @@ double PageViewItem::zoomFactor() const
     return m_zoomFactor;
 }
 
-void PageViewItem::setGeometry( int x, int y, int width, int height )
+qreal PageViewItem::absToPageX( qreal absX ) const
 {
-    m_geometry.setRect( x, y, width, height );
+    return ( absX - m_uncroppedGeometry.left() ) / m_uncroppedGeometry.width();
 }
 
-void PageViewItem::setWHZ( int w, int h, double z )
+qreal PageViewItem::absToPageY( qreal absY ) const
 {
-    m_geometry.setWidth( w );
-    m_geometry.setHeight( h );
+    return ( absY - m_uncroppedGeometry.top() ) / m_uncroppedGeometry.height();
+}
+
+bool PageViewItem::isVisible() const
+{
+    return m_visible;
+}
+
+QHash<int, FormWidgetIface*>& PageViewItem::formWidgets()
+{
+    return m_formWidgets;
+}
+
+void PageViewItem::setWHZC( int w, int h, double z, const Okular:: NormalizedRect & c )
+{
+    m_croppedGeometry.setWidth( w );
+    m_croppedGeometry.setHeight( h );
     m_zoomFactor = z;
+    m_crop = c;
+    QRect scaledCrop = c.geometry( w, h );
+    m_uncroppedGeometry.setWidth( qRound( w / ( c.right - c.left ) ) );
+    m_uncroppedGeometry.setHeight( qRound( h / ( c.bottom - c.top ) ) );
+    foreach(FormWidgetIface *fwi, m_formWidgets)
+    {
+        Okular::NormalizedRect r = fwi->rect();
+        fwi->setWidthHeight(
+            qRound( fabs( r.right - r.left ) * m_uncroppedGeometry.width() ),
+            qRound( fabs( r.bottom - r.top ) * m_uncroppedGeometry.height() ) );
+    }
 }
 
 void PageViewItem::moveTo( int x, int y )
+// Assumes setWHZC() has already been called
 {
-    m_geometry.moveLeft( x );
-    m_geometry.moveTop( y );
+    m_croppedGeometry.moveLeft( x );
+    m_croppedGeometry.moveTop( y );
+    m_uncroppedGeometry.moveLeft( qRound( x - m_crop.left * m_uncroppedGeometry.width() ) );
+    m_uncroppedGeometry.moveTop( qRound( y - m_crop.top * m_uncroppedGeometry.height() ) );
+    QHash<int, FormWidgetIface*>::iterator it = m_formWidgets.begin(), itEnd = m_formWidgets.end();
+    for ( ; it != itEnd; ++it )
+    {
+        Okular::NormalizedRect r = (*it)->rect();
+        (*it)->moveTo(
+            qRound( x + m_uncroppedGeometry.width() * r.left ) + 1,
+            qRound( y + m_uncroppedGeometry.height() * r.top ) + 1 );
+    }
+}
+
+void PageViewItem::setVisible( bool visible )
+{
+    setFormWidgetsVisible( visible && m_formsVisible );
+    m_visible = visible;
 }
 
 void PageViewItem::invalidate()
 {
-    m_geometry.setRect( 0, 0, 0, 0 );
+    m_croppedGeometry.setRect( 0, 0, 0, 0 );
+    m_uncroppedGeometry.setRect( 0, 0, 0, 0 );
+}
+
+bool PageViewItem::setFormWidgetsVisible( bool visible )
+{
+    m_formsVisible = visible;
+
+    if ( !m_visible )
+        return false;
+
+    bool somehadfocus = false;
+    QHash<int, FormWidgetIface*>::iterator it = m_formWidgets.begin(), itEnd = m_formWidgets.end();
+    for ( ; it != itEnd; ++it )
+    {
+        bool hadfocus = (*it)->setVisibility( visible );
+        somehadfocus = somehadfocus || hadfocus;
+    }
+    return somehadfocus;
 }
 
 /*********************/
@@ -103,12 +195,11 @@ PageViewMessage::PageViewMessage( QWidget * parent )
 {
     setObjectName( "pageViewMessage" );
     setFocusPolicy( Qt::NoFocus );
-    setAttribute( Qt::WA_OpaquePaintEvent );
     QPalette pal = palette();
     pal.setColor( QPalette::Active, QPalette::Window, QApplication::palette().color( QPalette::Active, QPalette::Window ) );
     setPalette( pal );
     // if the layout is LtR, we can safely place it in the right position
-    if ( QApplication::isLeftToRight() )
+    if ( layoutDirection() == Qt::LeftToRight )
         move( 10, 10 );
     resize( 0, 0 );
     hide();
@@ -124,79 +215,50 @@ void PageViewMessage::display( const QString & message, Icon icon, int durationM
         return;
     }
 
+    // set text
+    m_message = message;
+
     // determine text rectangle
-    QRect textRect = fontMetrics().boundingRect( message );
+    QRect textRect = fontMetrics().boundingRect( m_message );
     textRect.translate( -textRect.left(), -textRect.top() );
     textRect.adjust( 0, 0, 2, 2 );
     int width = textRect.width(),
-        height = textRect.height(),
-        textXOffset = 0,
-        shadowOffset = message.isRightToLeft() ? -1 : 1;
+        height = textRect.height();
 
     // load icon (if set) and update geometry
-    QPixmap symbol;
+    m_symbol = QPixmap();
     if ( icon != None )
     {
         switch ( icon )
         {
             case Annotation:
-                symbol = SmallIcon( "pencil" );
+                m_symbol = SmallIcon( "draw-freehand" );
                 break;
             case Find:
-                symbol = SmallIcon( "viewmag" );
+                m_symbol = SmallIcon( "zoom-original" );
                 break;
             case Error:
-                symbol = SmallIcon( "messagebox_critical" );
+                m_symbol = SmallIcon( "dialog-error" );
                 break;
             case Warning:
-                symbol = SmallIcon( "messagebox_warning" );
+                m_symbol = SmallIcon( "dialog-warning" );
                 break;
             default:
-                symbol = SmallIcon( "messagebox_info" );
+                m_symbol = SmallIcon( "dialog-information" );
                 break;
         }
-        textXOffset = 2 + symbol.width();
-        width += textXOffset;
-        height = qMax( height, symbol.height() );
+
+        width += 2 + m_symbol.width();
+        height = qMax( height, m_symbol.height() );
     }
-    QRect geometry( 0, 0, width + 10, height + 8 );
 
-    // resize pixmap, mask and widget
-    static QBitmap mask;
-    mask = QBitmap( geometry.size() + QSize( 1, 1 ) );
-    m_pixmap = QPixmap( geometry.size() + QSize( 1, 1 ) );
-    resize( geometry.size() + QSize( 1, 1 ) );
-
-    // create and set transparency mask
-    QPainter maskPainter( &mask);
-    mask.fill( Qt::white );
-    maskPainter.setBrush( Qt::black );
-    maskPainter.drawRoundRect( geometry, 1600 / geometry.width(), 1600 / geometry.height() );
-    maskPainter.end();
-    setMask( mask );
-
-    // draw background
-    QPalette pal = palette();
-    QPainter bufferPainter( &m_pixmap );
-    bufferPainter.setPen( Qt::black );
-    bufferPainter.setBrush( pal.color( QPalette::Window ) );
-    bufferPainter.drawRoundRect( geometry, 1600 / geometry.width(), 1600 / geometry.height() );
-
-    // draw icon if present
-    if ( !symbol.isNull() )
-        bufferPainter.drawPixmap( 5, 4, symbol, 0, 0, symbol.width(), symbol.height() );
-
-    // draw shadow and text
-    int yText = geometry.height() - height / 2;
-    bufferPainter.setPen( pal.color( QPalette::Window ).dark( 115 ) );
-    bufferPainter.drawText( 5 + textXOffset + shadowOffset, yText + 1, message );
-    bufferPainter.setPen( pal.color( QPalette::WindowText ) );
-    bufferPainter.drawText( 5 + textXOffset, yText, message );
+    // resize widget
+    resize( QRect( 0, 0, width + 10, height + 8 ).size() );
 
     // if the layout is RtL, we can move it to the right place only after we
     // know how much size it will take
-    if ( QApplication::isRightToLeft() )
-        move( parentWidget()->width() - geometry.width() - 10 - 1, 10 );
+    if ( layoutDirection() == Qt::RightToLeft )
+        move( parentWidget()->width() - geometry().width() - 10 - 1, 10 );
 
     // show widget and schedule a repaint
     show();
@@ -216,11 +278,40 @@ void PageViewMessage::display( const QString & message, Icon icon, int durationM
         m_timer->stop();
 }
 
-void PageViewMessage::paintEvent( QPaintEvent * e )
+void PageViewMessage::paintEvent( QPaintEvent * /* e */ )
 {
-    // paint the internal pixmap over the widget
-    QPainter p( this );
-    p.drawPixmap( e->rect().topLeft(), m_pixmap, e->rect() );
+    QRect textRect = fontMetrics().boundingRect( m_message );
+    textRect.translate( -textRect.left(), -textRect.top() );
+    textRect.adjust( 0, 0, 2, 2 );
+
+    int textXOffset = 0,
+        textYOffset = geometry().height() - textRect.height() / 2,
+        iconXOffset = 0,
+        iconYOffset = !m_symbol.isNull() ? ( geometry().height() - m_symbol.height() ) / 2 : 0,
+        shadowOffset = 1;
+
+    if ( layoutDirection() == Qt::RightToLeft )
+        iconXOffset = 2 + textRect.width();
+    else
+        textXOffset = 2 + m_symbol.width();
+
+    // draw background
+    QPainter painter( this );
+    painter.setRenderHint( QPainter::Antialiasing, true );
+    painter.setPen( Qt::black );
+    painter.setBrush( palette().color( QPalette::Window ) );
+    painter.translate( 0.5, 0.5 );
+    painter.drawRoundRect( 1, 1, width()-2, height()-2, 1600 / width(), 1600 / height() );
+
+    // draw icon if present
+    if ( !m_symbol.isNull() )
+        painter.drawPixmap( 5 + iconXOffset, iconYOffset, m_symbol, 0, 0, m_symbol.width(), m_symbol.height() );
+
+    // draw shadow and text
+    painter.setPen( palette().color( QPalette::Window ).dark( 115 ) );
+    painter.drawText( 5 + textXOffset + shadowOffset, textYOffset + shadowOffset, m_message );
+    painter.setPen( palette().color( QPalette::WindowText ) );
+    painter.drawText( 5 + textXOffset, textYOffset, m_message );
 }
 
 void PageViewMessage::mousePressEvent( QMouseEvent * /*e*/ )
@@ -240,9 +331,8 @@ PageViewTopMessage::PageViewTopMessage( QWidget * parent )
 {
     setAutoFillBackground( true );
     QPalette pal = palette();
-    pal.setColor( QPalette::Active, QPalette::Window, Qt::yellow );
-    pal.setColor( QPalette::Inactive, QPalette::Window, Qt::yellow );
-    pal.setColor( QPalette::Disabled, QPalette::Window, Qt::yellow );
+    KColorScheme::adjustBackground( pal, KColorScheme::NeutralBackground, QPalette::Window );
+    KColorScheme::adjustForeground( pal, KColorScheme::NeutralText, QPalette::Text );
     setPalette( pal );
     QHBoxLayout * lay = new QHBoxLayout( this );
     lay->setMargin( 4 );
@@ -253,6 +343,9 @@ PageViewTopMessage::PageViewTopMessage( QWidget * parent )
     lay->addWidget( m_label );
     m_label->setWordWrap( true );
     connect( m_label, SIGNAL( linkActivated( const QString& ) ), this, SIGNAL( action() ) );
+    m_button = new QToolButton( this );
+    m_button->setVisible( false );
+    lay->addWidget( m_button );
     QToolButton * closeButton = new QToolButton( this );
     lay->addWidget( closeButton );
     closeButton->setAutoRaise( true );
@@ -264,7 +357,7 @@ PageViewTopMessage::PageViewTopMessage( QWidget * parent )
     hide();
 }
 
-void PageViewTopMessage::display( const QString & message, const KIcon& icon )
+void PageViewTopMessage::setup( const QString & message, const KIcon& icon )
 {
     m_label->setText( message );
     resize( minimumSizeHint() );
@@ -276,7 +369,6 @@ void PageViewTopMessage::display( const QString & message, const KIcon& icon )
     {
         m_icon->setPixmap( icon.pixmap( m_icon->size() ) );
     }
-    show();
 }
 
 void PageViewTopMessage::setIconSize( int size )
@@ -284,32 +376,24 @@ void PageViewTopMessage::setIconSize( int size )
     m_icon->resize( size, size );
 }
 
+void PageViewTopMessage::setActionButton( QAction * action )
+{
+    m_button->setDefaultAction( action );
+    m_button->setVisible( action != 0 );
+}
 
 /*********************/
 /** PageViewToolBar  */
 /*********************/
 
-class ToolBarButton : public QToolButton
-{
-    public:
-        static const int iconSize = 32;
-        static const int buttonSize = 40;
-
-        ToolBarButton( QWidget * parent, const ToolBarItem & item );
-        int buttonID() const { return m_id; }
-
-    private:
-        int m_id;
-};
-
-ToolBarButton::ToolBarButton( QWidget * parent, const ToolBarItem & item )
-    : QToolButton( parent ), m_id( item.id )
+ToolBarButton::ToolBarButton( QWidget * parent, const AnnotationToolItem &item )
+    : QToolButton( parent ), m_id( item.id ), m_isText( item.isText )
 {
     setCheckable( true );
     setAutoRaise( true );
     resize( buttonSize, buttonSize );
     setIconSize( QSize( iconSize, iconSize ) );
-    setIcon( KIcon( item.pixmap ) );
+    setIcon( KIcon( item.pixmap, GuiUtils::iconLoader() ) );
     // set shortcut if defined
     if ( !item.shortcut.isEmpty() )
         setShortcut( QKeySequence( item.shortcut ) );
@@ -327,10 +411,25 @@ ToolBarButton::ToolBarButton( QWidget * parent, const ToolBarItem & item )
 /* PageViewToolBar */
 
 static const int toolBarGridSize = 40;
-static const int toolBarRBMargin = 2;
 
-struct ToolBarPrivate
+class ToolBarPrivate
 {
+public:
+    ToolBarPrivate( PageViewToolBar * qq )
+        : q( qq )
+    {
+    }
+
+    // rebuild contents and reposition then widget
+    void buildToolBar();
+    void reposition();
+    // compute the visible and hidden positions along current side
+    QPoint getInnerPoint() const;
+    QPoint getOuterPoint() const;
+    void selectButton( ToolBarButton * button );
+
+    PageViewToolBar * q;
+
     // anchored widget and side
     QWidget * anchorWidget;
     PageViewToolBar::Side anchorSide;
@@ -340,6 +439,7 @@ struct ToolBarPrivate
     QPoint currentPosition;
     QPoint endPosition;
     bool hiding;
+    bool visible;
 
     // background pixmap and buttons
     QPixmap backgroundPixmap;
@@ -347,15 +447,13 @@ struct ToolBarPrivate
 };
 
 PageViewToolBar::PageViewToolBar( QWidget * parent, QWidget * anchorWidget )
-    : QWidget( parent ), d( new ToolBarPrivate )
+    : QWidget( parent ), d( new ToolBarPrivate( this ) )
 {
-    setAttribute( Qt::WA_OpaquePaintEvent, true );
-    setMouseTracking( true );
-
     // initialize values of the private data storage structure
     d->anchorWidget = anchorWidget;
     d->anchorSide = Left;
     d->hiding = false;
+    d->visible = false;
 
     // create the animation timer
     d->animTimer = new QTimer( this );
@@ -371,12 +469,8 @@ PageViewToolBar::~PageViewToolBar()
     delete d;
 }
 
-void PageViewToolBar::showItems( Side side, const QLinkedList<ToolBarItem> & items )
+void PageViewToolBar::setItems( const QLinkedList<AnnotationToolItem> &items )
 {
-    // set parameters for sliding in
-    d->anchorSide = side;
-    d->hiding = false;
-
     // delete buttons if already present
     if ( !d->buttons.isEmpty() )
     {
@@ -387,7 +481,7 @@ void PageViewToolBar::showItems( Side side, const QLinkedList<ToolBarItem> & ite
     }
 
     // create new buttons for given items
-    QLinkedList<ToolBarItem>::const_iterator it = items.begin(), end = items.end();
+    QLinkedList<AnnotationToolItem>::const_iterator it = items.begin(), end = items.end();
     for ( ; it != end; ++it )
     {
         ToolBarButton * button = new ToolBarButton( this, *it );
@@ -396,24 +490,69 @@ void PageViewToolBar::showItems( Side side, const QLinkedList<ToolBarItem> & ite
     }
 
     // rebuild toolbar shape and contents
-    buildToolBar();
-    d->currentPosition = getOuterPoint();
-    d->endPosition = getInnerPoint();
-    move( d->currentPosition );
+    d->reposition();
+}
+
+void PageViewToolBar::setSide( Side side )
+{
+    d->anchorSide = side;
+
+    d->reposition();
+}
+
+void PageViewToolBar::showAndAnimate()
+{
+    // set parameters for sliding in
+    d->hiding = false;
+
     show();
 
+#ifdef OKULAR_ANIMATE_REVIEW_TOOBAR
     // start scrolling in
     d->animTimer->start( 20 );
+#else
+    d->currentPosition = d->endPosition;
+
+    move( d->currentPosition );
+
+    d->visible = true;
+#endif
 }
 
 void PageViewToolBar::hideAndDestroy()
 {
     // set parameters for sliding out
     d->hiding = true;
-    d->endPosition = getOuterPoint();
+    d->endPosition = d->getOuterPoint();
 
+#ifdef OKULAR_ANIMATE_REVIEW_TOOBAR
     // start scrolling out
     d->animTimer->start( 20 );
+#else
+    d->currentPosition = d->endPosition;
+
+    move( d->currentPosition );
+
+    d->visible = false;
+    deleteLater();
+#endif
+}
+
+void PageViewToolBar::selectButton( int id )
+{
+    ToolBarButton * button = 0;
+    if ( id >= 0 && id < d->buttons.count() )
+        button = *(d->buttons.begin() + id);
+    else
+    {
+        QLinkedList< ToolBarButton * >::const_iterator it = d->buttons.begin(), end = d->buttons.end();
+        for ( ; !button && it != end; ++it )
+            if ( (*it)->isChecked() )
+                button = *it;
+        if ( button )
+            button->setChecked( false );
+    }
+    d->selectButton( button );
 }
 
 bool PageViewToolBar::eventFilter( QObject * obj, QEvent * e )
@@ -425,7 +564,7 @@ bool PageViewToolBar::eventFilter( QObject * obj, QEvent * e )
         if ( d->hiding )
             deleteLater();
         else
-            reposition();
+            d->reposition();
     }
 
     // don't block event
@@ -436,7 +575,7 @@ void PageViewToolBar::paintEvent( QPaintEvent * e )
 {
     // paint the internal pixmap over the widget
     QPainter p( this );
-    p.drawPixmap( e->rect().topLeft(), d->backgroundPixmap, e->rect() );
+    p.drawImage( e->rect().topLeft(), d->backgroundPixmap.toImage(), e->rect() );
 }
 
 void PageViewToolBar::mousePressEvent( QMouseEvent * e )
@@ -466,7 +605,7 @@ void PageViewToolBar::mouseMoveEvent( QMouseEvent * e )
         return;
 
     d->anchorSide = side;
-    reposition();
+    d->reposition();
     emit orientationChanged( (int)side );
 }
 
@@ -477,17 +616,17 @@ void PageViewToolBar::mouseReleaseEvent( QMouseEvent * e )
         setCursor( Qt::ArrowCursor );
 }
 
-void PageViewToolBar::buildToolBar()
+void ToolBarPrivate::buildToolBar()
 {
-    int buttonsNumber = d->buttons.count(),
-        parentWidth = d->anchorWidget->width(),
-        parentHeight = d->anchorWidget->height(),
+    int buttonsNumber = buttons.count(),
+        parentWidth = anchorWidget->width(),
+        parentHeight = anchorWidget->height(),
         myCols = 1,
         myRows = 1;
 
     // 1. find out columns and rows we're going to use
-    bool topLeft = d->anchorSide == Left || d->anchorSide == Top;
-    bool vertical = d->anchorSide == Left || d->anchorSide == Right;
+    bool topLeft = anchorSide == PageViewToolBar::Left || anchorSide == PageViewToolBar::Top;
+    bool vertical = anchorSide == PageViewToolBar::Left || anchorSide == PageViewToolBar::Right;
     if ( vertical )
     {
         myCols = 1 + (buttonsNumber * toolBarGridSize) /
@@ -512,7 +651,7 @@ void PageViewToolBar::buildToolBar()
         myHeight += 16;
         myWidth += 4;
         yOffset += 12;
-        if ( d->anchorSide == Right )
+        if ( anchorSide == PageViewToolBar::Right )
             xOffset += 4;
     }
     else
@@ -520,50 +659,67 @@ void PageViewToolBar::buildToolBar()
         myWidth += 16;
         myHeight += 4;
         xOffset += 12;
-        if ( d->anchorSide == Bottom )
+        if ( anchorSide == PageViewToolBar::Bottom )
             yOffset += 4;
     }
 
-    bool prevUpdates = updatesEnabled();
-    setUpdatesEnabled( false );
+    bool prevUpdates = q->updatesEnabled();
+    q->setUpdatesEnabled( false );
 
     // 3. resize pixmap, mask and widget
-    static QBitmap mask;
-    mask = QBitmap( myWidth + 1, myHeight + 1 );
-    d->backgroundPixmap = QPixmap( myWidth + 1, myHeight + 1 );
-    resize( myWidth + 1, myHeight + 1 );
+    QBitmap mask( myWidth + 1, myHeight + 1 );
+    backgroundPixmap = QPixmap( myWidth + 1, myHeight + 1 );
+    backgroundPixmap.fill(Qt::transparent);
+    q->resize( myWidth + 1, myHeight + 1 );
 
-    // 4. create and set transparency mask
+    // 4. create and set transparency mask          // 4. draw background
     QPainter maskPainter( &mask);
-    mask.fill( Qt::white );
-    maskPainter.setBrush( Qt::black );
-    if ( vertical )
-        maskPainter.drawRoundRect( topLeft ? -10 : 0, 0, myWidth + 10, myHeight, 2000 / (myWidth + 10), 2000 / myHeight );
-    else
-        maskPainter.drawRoundRect( 0, topLeft ? -10 : 0, myWidth, myHeight + 10, 2000 / myWidth, 2000 / (myHeight + 10) );
-    setMask( mask );
+    mask.fill( Qt::white );     
+    maskPainter.setBrush( Qt::black );          
+    if ( vertical )     
+        maskPainter.drawRoundRect( topLeft ? -10 : 0, 0, myWidth + 11, myHeight, 2000 / (myWidth + 10), 2000 / myHeight );      
+    else        
+        maskPainter.drawRoundRect( 0, topLeft ? -10 : 0, myWidth, myHeight + 11, 2000 / myWidth, 2000 / (myHeight + 10) );      
+    maskPainter.end();          
+    q->setMask( mask );
 
     // 5. draw background
-    QPainter bufferPainter( &d->backgroundPixmap );
-    QPalette pal = palette();
+    QPainter bufferPainter( &backgroundPixmap );
+    bufferPainter.translate( 0.5, 0.5 );
+    QPalette pal = q->palette();
     // 5.1. draw horizontal/vertical gradient
-    QColor fromColor = topLeft ? pal.color( QPalette::Active, QPalette::Button ) : pal.color( QPalette::Active, QPalette::Light );
-    QColor toColor = topLeft ? pal.color( QPalette::Active, QPalette::Light ) : pal.color( QPalette::Active, QPalette::Button );
-    QImage gradientPattern = KImageEffect::gradient(
-            vertical ? QSize( myWidth + 1, 1 ) : QSize( 1, myHeight + 1 ), fromColor, toColor,
-            vertical ? KImageEffect::HorizontalGradient : KImageEffect::VerticalGradient );
-    bufferPainter.drawTiledPixmap( 0, 0, myWidth + 1, myHeight + 1, QPixmap::fromImage(gradientPattern) );
+    QLinearGradient grad;
+    switch ( anchorSide )
+    {
+        case PageViewToolBar::Left:
+            grad = QLinearGradient( 0, 1, myWidth + 1, 1 );
+            break;
+        case PageViewToolBar::Right:
+            grad = QLinearGradient( myWidth + 1, 1, 0, 1 );
+            break;
+        case PageViewToolBar::Top:
+            grad = QLinearGradient( 1, 0, 1, myHeight + 1 );
+            break;
+        case PageViewToolBar::Bottom:
+            grad = QLinearGradient( 1, myHeight + 1, 0, 1 );
+            break;
+    }
+    grad.setColorAt( 0, pal.color( QPalette::Active, QPalette::Button ) );
+    grad.setColorAt( 1, pal.color( QPalette::Active, QPalette::Light ) );
+    bufferPainter.setBrush( QBrush( grad ) );
     // 5.2. draw rounded border
-    bufferPainter.setPen( pal.color( QPalette::Active, QPalette::Dark ) );
+    bufferPainter.setPen( pal.color( QPalette::Active, QPalette::Dark ).lighter( 140 ) );
+    bufferPainter.setRenderHints( QPainter::Antialiasing );
     if ( vertical )
         bufferPainter.drawRoundRect( topLeft ? -10 : 0, 0, myWidth + 10, myHeight, 2000 / (myWidth + 10), 2000 / myHeight );
     else
         bufferPainter.drawRoundRect( 0, topLeft ? -10 : 0, myWidth, myHeight + 10, 2000 / myWidth, 2000 / (myHeight + 10) );
     // 5.3. draw handle
+    bufferPainter.translate( -0.5, -0.5 );
     bufferPainter.setPen( pal.color( QPalette::Active, QPalette::Mid ) );
     if ( vertical )
     {
-        int dx = d->anchorSide == Left ? 2 : 4;
+        int dx = anchorSide == PageViewToolBar::Left ? 2 : 4;
         bufferPainter.drawLine( dx, 6, dx + myWidth - 8, 6 );
         bufferPainter.drawLine( dx, 9, dx + myWidth - 8, 9 );
         bufferPainter.setPen( pal.color( QPalette::Active, QPalette::Light ) );
@@ -572,23 +728,25 @@ void PageViewToolBar::buildToolBar()
     }
     else
     {
-        int dy = d->anchorSide == Top ? 2 : 4;
+        int dy = anchorSide == PageViewToolBar::Top ? 2 : 4;
         bufferPainter.drawLine( 6, dy, 6, dy + myHeight - 8 );
         bufferPainter.drawLine( 9, dy, 9, dy + myHeight - 8 );
         bufferPainter.setPen( pal.color( QPalette::Active, QPalette::Light ) );
         bufferPainter.drawLine( 7, dy + 1, 7, dy + myHeight - 7 );
         bufferPainter.drawLine( 10, dy + 1, 10, dy + myHeight - 7 );
     }
+    bufferPainter.end();
 
     // 6. reposition buttons (in rows/col grid)
     int gridX = 0,
         gridY = 0;
-    QLinkedList< ToolBarButton * >::const_iterator it = d->buttons.begin(), end = d->buttons.end();
+    QLinkedList< ToolBarButton * >::const_iterator it = buttons.begin(), end = buttons.end();
     for ( ; it != end; ++it )
     {
         ToolBarButton * button = *it;
         button->move( gridX * toolBarGridSize + xOffset,
                       gridY * toolBarGridSize + yOffset );
+        button->show();
         if ( ++gridX == myCols )
         {
             gridX = 0;
@@ -596,45 +754,74 @@ void PageViewToolBar::buildToolBar()
         }
     }
 
-    setUpdatesEnabled( prevUpdates );
+    q->setUpdatesEnabled( prevUpdates );
 }
 
-void PageViewToolBar::reposition()
+void ToolBarPrivate::reposition()
 {
     // note: hiding widget here will gives better gfx, but ends drag operation
     // rebuild widget and move it to its final place
     buildToolBar();
-    d->currentPosition = getInnerPoint();
-    move( d->currentPosition );
+    if ( !visible )
+    {
+        currentPosition = getOuterPoint();
+        endPosition = getInnerPoint();
+    }
+    else
+    {
+        currentPosition = getInnerPoint();
+        endPosition = getOuterPoint();
+    }
+    q->move( currentPosition );
 
     // repaint all buttons (to update background)
-    QLinkedList< ToolBarButton * >::const_iterator it = d->buttons.begin(), end = d->buttons.end();
+    QLinkedList< ToolBarButton * >::const_iterator it = buttons.begin(), end = buttons.end();
     for ( ; it != end; ++it )
         (*it)->update();
 }
 
-QPoint PageViewToolBar::getInnerPoint()
+QPoint ToolBarPrivate::getInnerPoint() const
 {
     // returns the final position of the widget
-    if ( d->anchorSide == Left )
-        return QPoint( 0, (d->anchorWidget->height() - height()) / 2 );
-    if ( d->anchorSide == Top )
-        return QPoint( (d->anchorWidget->width() - width()) / 2, 0 );
-    if ( d->anchorSide == Right )
-        return QPoint( d->anchorWidget->width() - width() + toolBarRBMargin, (d->anchorWidget->height() - height()) / 2 );
-    return QPoint( (d->anchorWidget->width() - width()) / 2, d->anchorWidget->height() - height() + toolBarRBMargin );
+    QPoint newPos;
+    switch ( anchorSide )
+    {
+        case PageViewToolBar::Left:
+            newPos = QPoint( 0, ( anchorWidget->height() - q->height() ) / 2 );
+            break;
+        case PageViewToolBar::Top:
+            newPos = QPoint( ( anchorWidget->width() - q->width() ) / 2, 0 );
+            break;
+        case PageViewToolBar::Right:
+            newPos = QPoint( anchorWidget->width() - q->width(), ( anchorWidget->height() - q->height() ) / 2 );
+            break;
+        case PageViewToolBar::Bottom:
+            newPos = QPoint( ( anchorWidget->width() - q->width()) / 2, anchorWidget->height() - q->height() );
+            break;
+    }
+    return newPos + anchorWidget->pos();
 }
 
-QPoint PageViewToolBar::getOuterPoint()
+QPoint ToolBarPrivate::getOuterPoint() const
 {
     // returns the point from which the transition starts
-    if ( d->anchorSide == Left )
-        return QPoint( -width(), (d->anchorWidget->height() - height()) / 2 );
-    if ( d->anchorSide == Top )
-        return QPoint( (d->anchorWidget->width() - width()) / 2, -height() );
-    if ( d->anchorSide == Right )
-        return QPoint( d->anchorWidget->width() + toolBarRBMargin, (d->anchorWidget->height() - height()) / 2 );
-    return QPoint( (d->anchorWidget->width() - width()) / 2, d->anchorWidget->height() + toolBarRBMargin );
+    QPoint newPos;
+    switch ( anchorSide )
+    {
+        case PageViewToolBar::Left:
+            newPos = QPoint( -q->width(), ( anchorWidget->height() - q->height() ) / 2 );
+            break;
+        case PageViewToolBar::Top:
+            newPos = QPoint( ( anchorWidget->width() - q->width() ) / 2, -q->height() );
+            break;
+        case PageViewToolBar::Right:
+            newPos = QPoint( anchorWidget->width(), ( anchorWidget->height() - q->height() ) / 2 );
+            break;
+        case PageViewToolBar::Bottom:
+            newPos = QPoint( ( anchorWidget->width() - q->width() ) / 2, anchorWidget->height() );
+            break;
+    }
+    return newPos + anchorWidget->pos();
 }
 
 void PageViewToolBar::slotAnimate()
@@ -655,23 +842,50 @@ void PageViewToolBar::slotAnimate()
     {
         d->animTimer->stop();
         if ( d->hiding )
+        {
+            d->visible = false;
             deleteLater();
+        }
+        else
+        {
+            d->visible = true;
+        }
     }
 }
 
 void PageViewToolBar::slotButtonClicked()
 {
     ToolBarButton * button = qobject_cast<ToolBarButton *>( sender() );
+    d->selectButton( button );
+}
+
+void ToolBarPrivate::selectButton( ToolBarButton * button )
+{
     if ( button )
     {
         // deselect other buttons
-        QLinkedList< ToolBarButton * >::const_iterator it = d->buttons.begin(), end = d->buttons.end();
+        QLinkedList< ToolBarButton * >::const_iterator it = buttons.begin(), end = buttons.end();
         for ( ; it != end; ++it )
             if ( *it != button )
                 (*it)->setChecked( false );
         // emit signal (-1 if button has been unselected)
-        emit toolSelected( button->isChecked() ? button->buttonID() : -1 );
+        emit q->toolSelected( button->isChecked() ? button->buttonID() : -1 );
     }
 }
 
+void PageViewToolBar::setToolsEnabled( bool on )
+{
+    QLinkedList< ToolBarButton * >::const_iterator it = d->buttons.begin(), end = d->buttons.end();
+    for ( ; it != end; ++it )
+        (*it)->setEnabled( on );
+}
+
+void PageViewToolBar::setTextToolsEnabled( bool on )
+{
+    QLinkedList< ToolBarButton * >::const_iterator it = d->buttons.begin(), end = d->buttons.end();
+    for ( ; it != end; ++it )
+        if ( (*it)->isText() )
+            (*it)->setEnabled( on );
+}
+
 #include "pageviewutils.moc"
